/*============================================================================*/
/* eCall ANSI C fixed-point reference source code                             */
/*                                                                            */
/* File:         ecall_fec.c                                                  */
/* Version:      8.6.0 (Rel8) / 9.4.0 (Rel9)                                  */
/* Date:         2011-02-08                                                   */
/* Description:  Turbo encoder and decoder                                    */
/*----------------------------------------------------------------------------*/

#include "modemx.h"


/*============================================================================*/
/* ENCODER FUNCTION: FecEncode                                                */
/*----------------------------------------------------------------------------*/
/* Description: encoding of MSD                                               */
/*                                                                            */
/* InOut:  Ord1* buffer  <-> takes info bits and returns coded bits           */
/*----------------------------------------------------------------------------*/

void FecEncode(Ord1 *buffer)
{
  /* I> in: 1120 | out: 1148 -> (input bits to reside at buffer start) */
  AttachCrc(buffer, &buffer[NRB_INFO_CRC]);

  /* -> in: 1148 | out: 1148 -> */
  Scramble(&buffer[NRB_INFO_CRC], buffer);

  /* -> in: 1148 | out: 3456 O> */
  EncodeTwo(buffer);
}

/*============================================================================*/
/* ENCODER FUNCTION: AttachCrc                                                */
/*----------------------------------------------------------------------------*/
/* Description: attaches CRC bits                                             */
/*                                                                            */
/* In:     const Ord1* infoBits     -> input information bits                 */
/* Out:    Ord1*       infoWithCrc  <- bits with CRC attached                 */
/*----------------------------------------------------------------------------*/

void AttachCrc(const Ord1 *infoBits, Ord1 *infoWithCrc)
{
  Int16 i, j;

  /* copy over */
  for (i = 0; i < NRB_INFO_CRC; i++) {
    if (i < NRB_INFO)
      infoWithCrc[i] = infoBits[i];
    else
      infoWithCrc[i] = 0;
  }
  /* generate parity part */
  for (i = 0; i < NRB_INFO; i++) {
    if (infoWithCrc[i] == 1) {
      for (j = i; j < NRB_CRC+1+i; j++)
        infoWithCrc[j] ^= crcPolynomial[j-i];
    }
  }
  /* revert info part */
  for (i = 0; i < NRB_INFO; i++)
    infoWithCrc[i] = infoBits[i];
}

/*============================================================================*/
/* ENCODER FUNCTION: Scramble                                                 */
/*----------------------------------------------------------------------------*/
/* Description: bit scrambling                                                */
/*                                                                            */
/* In:     const Ord1* in   -> non scrambled input bit sequence               */
/* Out:    Ord1*       out  <- scrambled output bit sequence                  */
/*----------------------------------------------------------------------------*/

void Scramble(const Ord1 *in, Ord1 *out)
{
  Int16 k;
  for (k = 0; k < NRB_INFO_CRC; k++) {
    out[k] = in[k]^scramblingSeq[k];
  }
}

/*============================================================================*/
/* ENCODER FUNCTION: EncodeTwo                                                */
/*----------------------------------------------------------------------------*/
/* Description: encoding of bit sequence                                      */
/*                                                                            */
/* InOut:  Ord1* codedBits  <-> scrambled bits to coded bits                  */
/*----------------------------------------------------------------------------*/

void EncodeTwo(Ord1 *codedBits)
{
  Int16 i, j, k;
  Ord1 *input = codedBits;

  /* buffer structure: info1|itail1 itail2 parity1|ptail1 parity2|ptail2 */
  i = NRB_INFO_CRC + NRB_TAIL*2;
  j = NRB_INFO_CRC + NRB_TAIL + i;

  /* input sequence expected at buffer position 0 */
  for (k = 0; k < NRB_INFO_CRC; k++) {
    codedBits[i++] = input[k];
    codedBits[j++] = input[interleaverSeq[k]];
  }
  /* encoder one (info1: non-interleaved) */
  EncodeOne(codedBits, 1);

  /* encoder two (info2: interleaved)*/
  EncodeOne(codedBits, 2);
}

/*============================================================================*/
/* ENCODER FUNCTION: EncodeOne                                                */
/*----------------------------------------------------------------------------*/
/* Description: convolutional encoding of each component                      */
/*                                                                            */
/* In:     Int16 encNr       -> component number                              */
/* InOut:  Ord1* codedBits  <-> bits to be encoded                            */
/*----------------------------------------------------------------------------*/

void EncodeOne(Ord1 *codedBits, Int16 encNr)
{
  Int16 i, j, k;
  Ord16 encState, targetState;
  Ord1  encOut, tail;

  /* start parity(info) */
  i = NRB_INFO_CRC*(encNr) + NRB_TAIL*(encNr+1);
  encState = 0;

  /* input is overwritten by output */
  for (k = 0; k < NRB_INFO_CRC; k++) {
    encOut = outputParity[encState*2 + codedBits[i]];
    encState = stateTrans[encState*2 + codedBits[i]];
    codedBits[i++] = encOut;
  }
  /* start parity(tail) and tail */
  i = NRB_INFO_CRC*(encNr+1) + NRB_TAIL*(encNr+1);
  j = NRB_INFO_CRC+(encNr-1)*NRB_TAIL;

  /* terminate trellis */
  for (k = 0; k < NRB_TAIL; k++) {
    targetState = (encState << 1) & ((1 << NRB_TAIL)-1);
    tail = (stateTrans[encState*2] == targetState) ? 0 : 1;

    codedBits[i++] = outputParity[encState*2 + tail];
    codedBits[j++] = tail;

    encState = targetState;
  }
}

/*============================================================================*/
/* DECODER FUNCTION: FecDecode                                                */
/*----------------------------------------------------------------------------*/
/* Description: decoding to find the MSD                                      */
/*                                                                            */
/* In:     const IntLLR* in   -> received soft bits                           */
/*         Int16         rv   -> redundancy version                           */
/* Out:    Ord1*         out  <- decoded MSD in binary representation         */
/* Return: Bool               <- result of CRC check                          */
/*----------------------------------------------------------------------------*/

Bool FecDecode(const IntLLR *in, Int16 rv, Ord1 *out)
{
  Bool crcOk;
  static IntLLR chLLRbuffer[NRB_CODE_BUFFER];

  /* buffer is reinitialized every ARQ_MAX transmissions */
  if ((rv%ARQ_MAX) == 0) {
    memset(chLLRbuffer, 0, sizeof(IntLLR)*NRB_CODE_BUFFER);
  }
  /* update chLLRbuffer */
  UpdateBuffer(chLLRbuffer, in, rv);

  /* decode chLLRbuffer */
  DecodeBuffer(
    &chLLRbuffer[0],
    &chLLRbuffer[  NRB_INFO_CRC +   NRB_TAIL],
    &chLLRbuffer[  NRB_INFO_CRC + 2*NRB_TAIL],
    &chLLRbuffer[2*NRB_INFO_CRC + 3*NRB_TAIL], out);

  /* descramble decoded bits */
  Descramble(out);

  /* CRC check */
  crcOk = DecodeCrc(out);
  return crcOk;
}

/*============================================================================*/
/* DECODER FUNCTION: UpdateBuffer                                             */
/*----------------------------------------------------------------------------*/
/* Description: update channel LLR buffer with new soft bits                  */
/*                                                                            */
/* In:     const IntLLR* softInBits    -> received soft bits                  */
/*         Int16         rv            -> redundancy version                  */
/* InOut:  IntLLR*       chLLRbuffer  <-> decoder buffer                      */
/*----------------------------------------------------------------------------*/

void UpdateBuffer(IntLLR *chLLRbuffer, const IntLLR *softInBits, Int16 rv)
{
  Int32 temp;
  Int16 i, j = ((rv%2) ? 0 : 1)*(ARQ_MAX/2) + (rv/2)%(ARQ_MAX/2);

  for (i = 0; i < NRB_CODE_ARQ; i++) {
    temp = (Int32)chLLRbuffer[redVerIndex[j][i]] + (Int32)softInBits[i];

    chLLRbuffer[redVerIndex[j][i]] = (ABS(temp) < LLR_MAX) ?
      (IntLLR)temp : (IntLLR)(SIGN(temp)*LLR_MAX);
  }
}

/*============================================================================*/
/* DECODER FUNCTION: DecodeBuffer                                             */
/*----------------------------------------------------------------------------*/
/* Description: decoding of LLR buffer                                        */
/*                                                                            */
/* In:     const IntLLR* syst1    -> RX systematic soft bits                  */
/*         const IntLLR* syst2    -> interleaved RX systematic tail bits      */
/*         const IntLLR* parity1  -> RX parity soft bits                      */
/*         const IntLLR* parity2  -> interleaved RX parity soft bits          */
/* Out:    Ord1*         decBits  <- decoded bits                             */
/*----------------------------------------------------------------------------*/

void DecodeBuffer(const IntLLR *syst1, const IntLLR *syst2,
                  const IntLLR *parity1, const IntLLR *parity2, Ord1 *decBits)
{
  Int16 i, j;
  Int32 temp;
  IntLLR *Le12, *Le21;

  /* initialize memory */
  Le12 = (IntLLR*)&decBits[0];
  Le21 = (IntLLR*)&decBits[sizeof(IntLLR)*(NRB_INFO_CRC + NRB_TAIL)];

  memset(Le12, 0, sizeof(IntLLR)*(NRB_INFO_CRC + NRB_TAIL));
  memset(Le21, 0, sizeof(IntLLR)*(NRB_INFO_CRC + NRB_TAIL));

  /* iterative decoding */
  for (i = 0; i < FEC_ITERATIONS; i++) {
    memcpy(Le12, Le21, sizeof(IntLLR)*(NRB_INFO_CRC + NRB_TAIL));

    /* add received systematic bits to extrinsic information */
    for (j = 0; j < NRB_INFO_CRC+NRB_TAIL; j++) {
      temp = (Int32)Le12[j] + (Int32)syst1[j];
      Le12[j] = (ABS(temp) < LLR_MAX) ?
        (IntLLR)temp : (IntLLR)(SIGN(temp)*LLR_MAX);
    }
    /* decode code one (produces Le12) */
    Bcjr(parity1, Le12);

    /* interleave extrinsic information (produces interleaved Le12) */
    Interleave(Le12, Le21);

    /* add received systematic bits to extrinsic information */
    for (j = 0; j < NRB_INFO_CRC; j++) {
      temp = (Int32)Le21[j] + (Int32)syst1[interleaverSeq[j]];
      Le21[j] = (ABS(temp) < LLR_MAX) ?
        (IntLLR)temp : (IntLLR)((SIGN(temp))*LLR_MAX);
    }
    for (j = 0; j < NRB_TAIL; j++) {
      Le21[j+NRB_INFO_CRC] = syst2[j];
    }
    /* decode code two (produces interleaved Le21) */
    Bcjr(parity2, Le21);

    /* deinterleave extrinsic information (produces Le21) */
    Deinterleave(Le21);
  }

  /* take final decision on bits */
  for (j = 0; j < NRB_INFO_CRC; j++) {
    decBits[j] =
      ((Int32)syst1[j] + (Int32)Le21[j] + (Int32)Le12[j] > 0) ? 0 : 1;
  }
}

/*============================================================================*/
/* DECODER FUNCTION: Bcjr                                                     */
/*----------------------------------------------------------------------------*/
/* Description: BCJR algorithm                                                */
/*                                                                            */
/* In:     const IntLLR* parity      -> received parity soft bits             */
/* InOut:  IntLLR*       extrinsic  <-> extrinsic information                 */
/*----------------------------------------------------------------------------*/

void Bcjr(const IntLLR *parity, IntLLR *extrinsic)
{
  Int16  i, k;
  Int32  temp;
  IntLLR nom, den, norm, g1, g2, *alpha1, *alpha2, *bTemp1, *bTemp2;

  IntLLR alphaQ[2*FEC_STATES];
  IntLLR betaQ[(NRB_INFO_CRC + NRB_TAIL + 1)*FEC_STATES];

  /* initialize betaQ */
  bTemp1 = &betaQ[(NRB_INFO_CRC+NRB_TAIL)*FEC_STATES];
  bTemp1[0] = 0;
  for (i = 1; i < FEC_STATES; i++) {
    bTemp1[i] = -LLR_MAX;
  }
  /* compute betaQ (backward iteration) */
  for (k = NRB_INFO_CRC+NRB_TAIL; k >= 1; k--) {
    bTemp1 = &betaQ[(k-1)*FEC_STATES];
    bTemp2 = &bTemp1[FEC_STATES];
    norm   = -LLR_MAX;

    for (i = 0; i < FEC_STATES; i++) {
      g1 = GammaQ(k, (i<<1), parity, extrinsic);
      g2 = GammaQ(k, (i<<1)+1, parity, extrinsic);

      bTemp1[i] = JacLog(
        (Int32)bTemp2[stateTrans[i<<1]] + (Int32)g1,
        (Int32)bTemp2[stateTrans[(i<<1)+1]] + (Int32)g2);
      norm = JacLog(
        (Int32)bTemp1[i], (Int32)norm);
    }
    /* normalization of betaQ */
    for (i = 0; i < FEC_STATES; i++) {
      temp = (Int32)bTemp1[i] - (Int32)norm;
      bTemp1[i] = (temp < (-LLR_MAX)) ? (IntLLR)(-LLR_MAX) : (IntLLR)temp;
    }
  }

  /* initialize alphaQ */
  alpha2 = &alphaQ[FEC_STATES];
  alpha1 = &alphaQ[0];
  alpha1[0] = 0;
  for (i = 1; i < FEC_STATES; i++) {
    alpha1[i] = -LLR_MAX;
  }
  /* compute alphaQ (forward iteration) */
  for (k = 1; k <= NRB_INFO_CRC+NRB_TAIL; k++) {
    if (k-1 < NRB_INFO_CRC) {
      bTemp1 = &betaQ[k*FEC_STATES];
      norm   = -LLR_MAX;

      for (i = 0; i < FEC_STATES; i++) {
        g1 = GammaQ(k, (revStateTrans[i<<1] << 1), parity, extrinsic);
        g2 = GammaQ(k, (revStateTrans[(i<<1)+1] << 1)+1, parity, extrinsic);

        alpha2[i] = JacLog(
          (Int32)alpha1[revStateTrans[i<<1]] + (Int32)g1,
          (Int32)alpha1[revStateTrans[(i<<1)+1]] + (Int32)g2);
        norm = JacLog(
          (Int32)alpha2[i], (Int32)norm);
      }
      /* normalization of alphaQ */
      for (i = 0; i < FEC_STATES; i++) {
        temp = (Int32)alpha2[i] - (Int32)norm;
        alpha2[i] = (temp < (-LLR_MAX)) ? (IntLLR)(-LLR_MAX) : (IntLLR)temp;
      }
      /* calculate extrinsic output */
      nom = -LLR_MAX;
      den = -LLR_MAX;

      for (i = 0; i < FEC_STATES; i++) {
        if (outputParity[i<<1] == 0) {
          nom = JacLog(
            (Int32)nom,
            (Int32)alpha1[i] +
            (Int32)bTemp1[stateTrans[i<<1]] +
            (Int32)parity[k-1]/2);
        } else {
          nom = JacLog(
            (Int32)nom,
            (Int32)alpha1[i] +
            (Int32)bTemp1[stateTrans[i<<1]] -
            (Int32)parity[k-1]/2);
        }
        if (outputParity[(i<<1)+1] == 0) {
          den = JacLog(
            (Int32)den,
            (Int32)alpha1[i] +
            (Int32)bTemp1[stateTrans[(i<<1)+1]] +
            (Int32)parity[k-1]/2);
        } else {
          den = JacLog(
            (Int32)den,
            (Int32)alpha1[i] +
            (Int32)bTemp1[stateTrans[(i<<1)+1]] -
            (Int32)parity[k-1]/2);
        }
      }
      temp = (Int32)nom - (Int32)den;
      extrinsic[k-1] = (ABS(temp) < LLR_MAX) ?
        (IntLLR)temp : (IntLLR)(SIGN(temp)*LLR_MAX);

      /* swap alphaQ states */
      alpha1 = &alphaQ[FEC_STATES*(k%2)];
      alpha2 = &alphaQ[FEC_STATES*((k-1)%2)];
    } else {
      extrinsic[k-1] = 0;
    }
  }
}

/*============================================================================*/
/* DECODER FUNCTION: Interleave                                               */
/*----------------------------------------------------------------------------*/
/* Description: Turbo code interleaver                                        */
/*                                                                            */
/* In:     const IntLLR* in   -> input sequence                               */
/* Out:    IntLLR*       out  <- output sequence                              */
/*----------------------------------------------------------------------------*/

void Interleave(const IntLLR *in, IntLLR *out)
{
  Int16 i;
  for (i = 0; i < NRB_INFO_CRC; i++) {
    out[i] = in[interleaverSeq[i]];
  }
}

/*============================================================================*/
/* DECODER FUNCTION: Deinterleave                                             */
/*----------------------------------------------------------------------------*/
/* Description: Turbo code deinterleaver                                      */
/*                                                                            */
/* InOut:  IntLLR* inout  <-> input and deinterleaved output sequence         */
/*----------------------------------------------------------------------------*/

void Deinterleave(IntLLR *inout)
{
  Int16 i;
  IntLLR temp[NRB_INFO_CRC];

  for (i = 0; i < NRB_INFO_CRC; i++) {
    temp[interleaverSeq[i]] = inout[i];
  }
  memcpy(inout, temp, sizeof(IntLLR)*NRB_INFO_CRC);
}

/*============================================================================*/
/* DECODER FUNCTION: Descramble                                               */
/*----------------------------------------------------------------------------*/
/* Description: descrambles decoded bits                                      */
/*                                                                            */
/* InOut:  Ord1* inout  <-> input and output bit sequence                     */
/*----------------------------------------------------------------------------*/

void Descramble(Ord1 *inout)
{
  Int16 k;
  Ord1 temp[NRB_INFO_CRC];

  for (k = 0; k < NRB_INFO_CRC; k++) {
    temp[k] = inout[k]^scramblingSeq[k];
  }
  memcpy(inout, temp, sizeof(Ord1)*NRB_INFO_CRC);
}

/*============================================================================*/
/* DECODER FUNCTION: DecodeCrc                                                */
/*----------------------------------------------------------------------------*/
/* Description: check CRC of decoded bits                                     */
/*                                                                            */
/* In:     const Ord1* codedBits  -> decoded bit sequence to be checked       */
/* Return: Bool                   <- result of CRC check                      */
/*----------------------------------------------------------------------------*/

Bool DecodeCrc(const Ord1 *codedBits)
{
  Int16 i, j;
  Bool crcOk = True;
  Ord1 temp[NRB_INFO_CRC];

  memcpy(temp, codedBits, sizeof(Ord1)*NRB_INFO_CRC);

  for (i = 0; i < NRB_INFO; i++) {
    if (temp[i] == 1)
      for (j = i; j < NRB_CRC+1+i; j++)
        temp[j] ^= crcPolynomial[j-i];
  }
  for (i = NRB_INFO; i < NRB_INFO_CRC; i++) {
    if (temp[i] != 0) {
      crcOk = False;
      break;
    }
  }
  return crcOk;
}

/*============================================================================*/
/* DECODER FUNCTION: GammaQ                                                   */
/*----------------------------------------------------------------------------*/
/* Description: compute gamma values for BCJR algorithm                       */
/*                                                                            */
/* In:     Int16         k          -> bit position                           */
/*         Int16         l          -> state                                  */
/*         const IntLLR* parity     -> received parity bits                   */
/*         const IntLLR* extrinsic  -> sum of extrinsic and systematic bits   */
/* Return: IntLLR                   <- value of gamma(k,l)                    */
/*----------------------------------------------------------------------------*/

IntLLR GammaQ(Int16 k, Int16 l, const IntLLR *parity, const IntLLR *extrinsic)
{
  /* extrinsic must include received systematic bits */
  if (l%2) {
    if (outputParity[l] == 0) {
      return (IntLLR)((-(Int32)extrinsic[k-1] + (Int32)parity[k-1])/2);
    } else {
      return (IntLLR)((-(Int32)extrinsic[k-1] - (Int32)parity[k-1])/2);
    }
  } else {
    if (outputParity[l] == 0) {
      return (IntLLR)(((Int32)extrinsic[k-1] + (Int32)parity[k-1])/2);
    } else {
      return (IntLLR)(((Int32)extrinsic[k-1] - (Int32)parity[k-1])/2);
    }
  }
}

/*============================================================================*/
/* UTILITY FUNCTION: JacLog                                                   */
/*----------------------------------------------------------------------------*/
/* Description: Jacobian logarithm                                            */
/*                                                                            */
/* In:     IntLLR a  -> value one                                             */
/*         IntLLR b  -> value two                                             */
/* Return: IntLLR    <- Jacobian logarithm                                    */
/*----------------------------------------------------------------------------*/

IntLLR JacLog(Int32 a, Int32 b)
{
  Int32 maxab, ind;

  if (a > b) {
    maxab = a;
    ind = (a-b) >> (LOGEXP_DELTA + LOGEXP_QIN);
  } else {
    maxab = b;
    ind = (b-a) >> (LOGEXP_DELTA + LOGEXP_QIN);
  }
  maxab += (ind >= LOGEXP_RES) ? 0 : logExpTable[ind];

  if (maxab >= LLR_MAX) {
    return LLR_MAX;
  } else {
    return (IntLLR)maxab;
  }
}
