/*============================================================================*/
/* eCall ANSI C fixed-point reference source code                             */
/*                                                                            */
/* File:         ecall_sync.c                                                 */
/* Version:      12.0.0 (Rel12)                                               */
/* Date:         2014-09-11                                                   */
/* Description:  synchronization functions                                    */
/*----------------------------------------------------------------------------*/

#include "modemx.h"


/*============================================================================*/
/* UTILITY FUNCTION: SyncReset                                                */
/*----------------------------------------------------------------------------*/
/* InOut:  SyncState* sync       <-> sync struct                              */
/* In:     Int32*     mem         -> pointer to sync memory                   */
/*         Int32*     memWakeup   -> pointer to sync wakeup memory            */
/*----------------------------------------------------------------------------*/

void SyncReset(SyncState *sync, Int32 *mem, Int32 *memWakeup)
{
  Int16 i;

  /* assign buffer pointers */
  if (mem != NULL) {
    sync->mem = mem;
    sync->memWakeup = memWakeup;
  }
  /* reset variables */
  sync->flag      = False;
  sync->invert    = Minus;
  sync->resync    = False;
  sync->delay     = 0;
  sync->delayMem  = 0;
  sync->npeaks    = 0;
  sync->npeaksMem = 0;
  sync->events    = 0;
  sync->sign      = 0;

  /* clear memory */
  for (i = 0; i <= 2*NRS_CP; i++) {
    sync->shape[i] = 0;
  }
  sync->amplitude[0] = 0;
  sync->amplitude[1] = 0;
  sync->amplitude[2] = 0;

  /* clear sync subsystem */
  SyncSubReset(&sync->syncPos);
  SyncSubReset(&sync->syncNeg);

  /* initialize control message sync */
  sync->check    = Zero;
  sync->checkCnt = 0;
  sync->index    = NRF_DLMUTE1LL + NRF_DLDATA - 1;

  /* clear sync buffers */
  if (sync->mem != NULL) {
    memset(sync->mem, 0, sizeof(Int32)*NRS_MEMSYNC);
  }
  if (sync->memWakeup != NULL) {
    memset(sync->memWakeup, 0, sizeof(Int32)*2*(NRF_SYNC+1));
  }
}

/*============================================================================*/
/* UTILITY FUNCTION: SyncSubReset                                             */
/*----------------------------------------------------------------------------*/
/* InOut:  SyncSub* ssub  <-> sync subsystem                                  */
/*----------------------------------------------------------------------------*/

void SyncSubReset(SyncSub *ssub)
{
  Int16 i;

  for (i = 0; i <= 2*NRS_CP; i++) {
    ssub->shape[i] = 0;
  }
  ssub->flag   = False;
  ssub->delay  = 0;
  ssub->npeaks = 0;

  ssub->amplitude[0] = 0;
  ssub->amplitude[1] = 0;
  ssub->sign = 0;
}

/*============================================================================*/
/* FUNCTION: Sync                                                             */
/*----------------------------------------------------------------------------*/
/* Description: main synchronization function                                 */
/*                                                                            */
/* InOut:  SyncState*   sync    <-> sync struct                               */
/* In:     const Int16* pcm      -> input frame                               */
/*         const char*  caller   -> modem identification                      */
/*         Bool         invert   -> port inversion flag                       */
/*----------------------------------------------------------------------------*/

void Sync(SyncState *sync, const Int16 *pcm, const char *caller, Bool invert)
{
  Int16 offset = syncIndexPreamble[SYNC_IDXLEN-1] + PCM_LENGTH + 2*NRS_CP;

  /* set buffer pointers */
  Int32 *posP  = &sync->mem[offset +  0];
  Int32 *posN  = &sync->mem[offset +  9];
  Int32 *corrP = &sync->mem[offset + 18];
  Int32 *corrN = &sync->mem[offset + 27 + 9*2*NRS_CP];

  /* test for 500 Hz or 800 Hz sine tone and run sync filter */
  ToneDetect(sync, pcm);
  SyncFilter(sync, pcm, invert);

  /* copy data to subsystem */
  SyncSubPut(sync, &sync->syncPos);
  SyncSubPut(sync, &sync->syncNeg);

  /* peak evaluation for each subsystem */
  SyncSubRun(&sync->syncPos, caller, posP, corrP, posN, corrN);
  SyncSubRun(&sync->syncNeg, caller, posN, corrN, posP, corrP);

  /* decision logic */
  if (sync->syncPos.flag == True) {
    if (
        sync->syncNeg.flag == True &&
        (sync->syncNeg.npeaks >= 4 || (sync->syncNeg.npeaks >= 3 && sync->syncPos.npeaks>0 )) &&
        sync->syncPos.sign < sync->syncNeg.sign &&
        sync->syncPos.sign <= 0 &&
        sync->syncNeg.sign > 0  
      )
    {
      sync->invert = True;
      sync->sign = sync->syncNeg.sign; 
      sync->npeaksMem = sync->syncNeg.npeaks;

      SyncSubGet(sync, &sync->syncNeg);

      LogInfo("[%-4s] sync detected; delay: %+4d; npeaks: %+4d; sign: %+4d, Pos.sign: %+4d, Neg.sign: %+4d (inverted sync)",
        caller, sync->delay, sync->npeaksMem, sync->sign, sync->syncPos.sign, sync->syncNeg.sign);
    }
    else if (
        sync->invert == False || (
          sync->syncPos.flag == True &&
          (sync->syncPos.npeaks >= 4 || (sync->syncPos.npeaks >= 3 && (sync->syncNeg.npeaks>0 || sync->syncNeg.flag == False))) && 
          sync->syncPos.sign > 0
      )) 
    {
      sync->invert = False;
      sync->sign = sync->syncPos.sign; 
      sync->npeaksMem = sync->syncPos.npeaks;

      SyncSubGet(sync, &sync->syncPos);

      LogInfo("[%-4s] sync detected; delay: %+4d; npeaks: %+4d; sign: %+4d, Pos.sign: %+4d, Neg.sign: %+4d (regular sync)",
        caller, sync->delay, sync->npeaksMem, sync->sign, sync->syncPos.sign, sync->syncNeg.sign);
    }
    else {
      sync->flag = False;
    }
  }
  else if (
      sync->syncNeg.flag == True && (
          sync->invert == True || (
            (sync->syncNeg.npeaks >= 4 || (sync->syncNeg.npeaks >= 3 && sync->syncPos.npeaks>0)) &&	        
          sync->syncNeg.sign > 0  
        ))
      ) 
    {
    sync->invert = True;
    sync->sign = sync->syncNeg.sign; 
    sync->npeaksMem = sync->syncNeg.npeaks;

    SyncSubGet(sync, &sync->syncNeg);

    LogInfo("[%-4s] sync detected; delay: %+4d; npeaks: %+4d; sign: %+4d, Pos.sign: %+4d, Neg.sign: %+4d (inverted sync)",
      caller, sync->delay, sync->npeaksMem, sync->sign, sync->syncPos.sign, sync->syncNeg.sign);
  }
  else {
    sync->flag = False;
  }
  if (sync->flag == True) {
    SyncSubReset(&sync->syncPos);
    SyncSubReset(&sync->syncNeg);
  }
}

/*============================================================================*/
/* IVS FUNCTION: SyncCheck                                                    */
/*----------------------------------------------------------------------------*/
/* Description: check whether locked sync is still valid                      */
/*                                                                            */
/* InOut:  SyncState*   sync    <-> sync struct                               */
/* In:     const Int16* pcm      -> input frame                               */
/*         const char*  caller   -> modem identification                      */
/*----------------------------------------------------------------------------*/

void SyncCheck(SyncState *sync, const Int16 *pcm, const char *caller)
{
  Int16 i, ref, dmax, dcen, dmin, imax, icen, imin;
  Int16 nCheck, maxCheck;
  Int16 offset = syncIndexPreamble[SYNC_IDXLEN-1] + PCM_LENGTH + 2*NRS_CP;

  Int16 delay  = 0;
  Int16 npeaks = 0;
  Int32 pulseP = 0;
  Int32 pulseN = 0;

  /* auxiliary struct */
  SyncSub ssub;

  /* set buffer pointers */
  Int32 *posP  = &sync->mem[offset +  0];
  Int32 *posN  = &sync->mem[offset +  9];
  Int32 *corrP = &sync->mem[offset + 18];
  Int32 *corrN = &sync->mem[offset + 27 + 9*2*NRS_CP];

  /* initialize and compute search window */
  sync->flag = False;

  if (sync->delayMem > 0 || sync->delayMem <= -PCM_LENGTH) {
    Abort("[%-4s] sync delay (%+1d) out of range!", caller, sync->delayMem);
  }
  if (NRS_CHECK > 480) {
    nCheck = 480;
    LogInfo("[%-4s] maximum sync checking range is [-480...480] samples",
      caller);
  } else {
    nCheck = NRS_CHECK;
  }

  ref  = sync->delayMem - nCheck + 4 + NRS_CP - PCM_LENGTH;
  imin = ref/PCM_LENGTH + 1 + NRF_DLCHUNK;
  dmin = ref%PCM_LENGTH - 4 - NRS_CP;

  ref  = sync->delayMem + 4 + NRS_CP - PCM_LENGTH;
  icen = ref/PCM_LENGTH + 1 + NRF_DLCHUNK;
  dcen = ref%PCM_LENGTH - 4 - NRS_CP;

  ref  = sync->delayMem + nCheck + 4 + NRS_CP + PCM_LENGTH;
  imax = (ref-1)/PCM_LENGTH + NRF_DLCHUNK;
  dmax = (ref-1)%PCM_LENGTH + 1 - PCM_LENGTH - 4 - NRS_CP;

  /* sync filter function */
  SyncFilter(sync, pcm, False);

  /* run sync check */
  if (sync->index == icen) {
    SyncSubChk(&sync->syncPos, caller,
      posP, corrP, posN, corrN, sync->delayMem);
    SyncSubChk(&sync->syncNeg, caller,
      posN, corrN, posP, corrP, sync->delayMem);
  }

  /* run sync tracking */
  if (sync->index >= imin && sync->index <= imax) {

    /* check for regular sync */
    sync->syncPos.delay -= PCM_LENGTH;

    SyncSubCpy(&sync->syncPos, &ssub);
    SyncSubReset(&sync->syncPos);

    sync->syncPos.amplitude[1] = sync->amplitude[1];
    SyncSubRun(&sync->syncPos, caller, posP, corrP, posN, corrN);

    if (sync->syncPos.flag == True && (
      (sync->index == imin && sync->syncPos.delay < dmin) ||
      (sync->index == imax && sync->syncPos.delay > dmax)))
    {
      sync->syncPos.flag = False;
    }
    if (ssub.flag == True && sync->syncPos.flag == True) {
      if (sync->syncPos.npeaks < ssub.npeaks ||
        (sync->syncPos.npeaks == ssub.npeaks &&
        sync->syncPos.amplitude[0] < ssub.amplitude[0]))
      {
        SyncSubCpy(&ssub, &sync->syncPos);
      }
    } else if (ssub.flag == True) {
      SyncSubCpy(&ssub, &sync->syncPos);
    }

    /* check for inverted sync */
    sync->syncNeg.delay -= PCM_LENGTH;

    SyncSubCpy(&sync->syncNeg, &ssub);
    SyncSubReset(&sync->syncNeg);

    sync->syncNeg.amplitude[1] = sync->amplitude[1];
    SyncSubRun(&sync->syncNeg, caller, posN, corrN, posP, corrP);

    if (sync->syncNeg.flag == True && (
      (sync->index == imin && sync->syncNeg.delay < dmin) ||
      (sync->index == imax && sync->syncNeg.delay > dmax)))
    {
      sync->syncNeg.flag = False;
    }
    if (ssub.flag == True && sync->syncNeg.flag == True) {
      if (sync->syncNeg.npeaks < ssub.npeaks ||
        (sync->syncNeg.npeaks == ssub.npeaks &&
        sync->syncNeg.amplitude[0] < ssub.amplitude[0]))
      {
        SyncSubCpy(&ssub, &sync->syncNeg);
      }
    } else if (ssub.flag == True) {
      SyncSubCpy(&ssub, &sync->syncNeg);
    }
  }

  /* take final sync check and tracking decision */
  if (sync->index == imax) {
    sync->flag = True;
    maxCheck = MAX(sync->syncPos.npeaksChk, sync->syncNeg.npeaksChk);

    if (sync->syncPos.flag == True && sync->syncNeg.flag == True) {
      for (i = 0; i <= 2*NRS_CP; i++) {
        pulseP += (sync->shape[i])*(sync->syncPos.shape[i]);
        pulseN += (sync->shape[i])*(sync->syncNeg.shape[i]);
      }
      pulseP = (pulseP >> 16)*(sync->syncPos.amplitude[0] >> 16);
      pulseN = (pulseN >> 16)*(sync->syncNeg.amplitude[0] >> 16);

      if ((
        sync->syncPos.npeaks >= sync->syncNeg.npeaks && pulseP > pulseN) || (
        sync->syncPos.npeaks > sync->syncNeg.npeaks && pulseP > pulseN/2 &&
        sync->syncNeg.npeaksChk == 0))
      {
        sync->check = Plus;
        delay  = sync->syncPos.delay;
        npeaks = sync->syncPos.npeaks;
      }
      else if ((
        sync->syncPos.npeaks <= sync->syncNeg.npeaks && pulseP < pulseN) || (
        sync->syncPos.npeaks < sync->syncNeg.npeaks && pulseP/2 < pulseN &&
        sync->syncPos.npeaksChk == 0))
      {
        sync->check = Minus;
        delay  = sync->syncNeg.delay;
        npeaks = sync->syncNeg.npeaks;
      }
      else if (sync->syncPos.npeaksChk > sync->syncNeg.npeaksChk) {
        sync->check = Plus;
      }
      else if (sync->syncNeg.npeaksChk > sync->syncPos.npeaksChk) {
        sync->check = Minus;
      }
      else {
        sync->check = Zero;
      }
    }
    else if (sync->syncPos.flag == True) {
      sync->check = Plus;
      delay  = sync->syncPos.delay;
      npeaks = sync->syncPos.npeaks;
    }
    else if (sync->syncNeg.flag == True) {
      sync->check = Minus;
      delay  = sync->syncNeg.delay;
      npeaks = sync->syncNeg.npeaks;
    }
    else if (sync->syncPos.npeaksChk > sync->syncNeg.npeaksChk) {
      sync->check = Plus;
    }
    else if (sync->syncNeg.npeaksChk > sync->syncPos.npeaksChk) {
      sync->check = Minus;
    }
    else {
      sync->check = Zero;
    }

    if (delay != 0 &&
      delay != (i = (sync->delayMem)-(imax%NRF_DLCHUNK)*PCM_LENGTH))
    {
      if (ABS(delay-i) == 2 || maxCheck < npeaks) {
        sync->delayMem = delay + (imax%NRF_DLCHUNK)*PCM_LENGTH;

        LogInfo("[%-4s] sync check detects sample slip! [%+1d sample(s)]",
          caller, delay-i);
      }
      else if (sync->syncPos.npeaksChk > sync->syncNeg.npeaksChk) {
        sync->check = Plus;
      }
      else if (sync->syncNeg.npeaksChk > sync->syncPos.npeaksChk) {
        sync->check = Minus;
      }
      else {
        sync->check = Zero;
      }
    }
    SyncSubReset(&sync->syncPos);
    SyncSubReset(&sync->syncNeg);

    if (sync->mem != NULL) {
      memset(sync->mem, 0, sizeof(Int32)*NRS_MEMSYNC);
    }
  }
}

/*============================================================================*/
/* IVS FUNCTION: SyncTrack                                                    */
/*----------------------------------------------------------------------------*/
/* Description: uplink sync tracker                                           */
/*                                                                            */
/* InOut:  SyncState* sync    <-> sync struct                                 */
/* In:     Bool       invert   -> port inversion flag                         */
/*----------------------------------------------------------------------------*/

void SyncTrack(SyncState *sync, Bool invert)
{
  Int16 i, ref, dmin, dmax, imin, imax, nTrack, indP, indN;

  Int16 offset = syncIndexPreamble[SYNC_IDXLEN-1] + PCM_LENGTH + 2*NRS_CP;
  Int16 npulse = 2*NRS_CP+1;
  Int16 npeaks = 0;

  Int32 *posP, *posN, *corrP, *corrN, ampP, ampN;

  /* set buffer pointers */
  if (invert == True) {
    posN  = &sync->mem[offset +  0];
    posP  = &sync->mem[offset +  9];
    corrN = &sync->mem[offset + 18];
    corrP = &sync->mem[offset + 27 + 9*2*NRS_CP];
  } else {
    posP  = &sync->mem[offset +  0];
    posN  = &sync->mem[offset +  9];
    corrP = &sync->mem[offset + 18];
    corrN = &sync->mem[offset + 27 + 9*2*NRS_CP];
  }

  /* initialize and compute search window */
  sync->flag = False;

  if (sync->delay > 0 || sync->delay <= -PCM_LENGTH) {
    Abort("[PSAP] sync delay (%+1d) out of range!", sync->delayMem);
  }
  if (NRS_TRACK > 240) {
    nTrack = 240;
    LogInfo("[PSAP] maximum sync tracking range is [-240...240] samples");
  } else {
    nTrack = NRS_TRACK;
  }
  ref = sync->delay + 4 + PCM_LENGTH + NRS_CP;

  /* check peaks at expected position */
  indP = 7 + (i = ref - 2*PCM_LENGTH - PEAK_DIST_PN)/PCM_LENGTH;
  if (posP[indP] == i) {
    npeaks++;
  }
  indN = 7 + (i = ref - 2*PCM_LENGTH)/PCM_LENGTH;
  if (posN[indN] == i) {
    npeaks++;
  }

  /* search for positive peak */
  dmin = ref - 2*PCM_LENGTH - PEAK_DIST_PN - nTrack;
  dmax = ref - 2*PCM_LENGTH - PEAK_DIST_PN + nTrack;

  imin = 0;
  while (posP[imin] < dmin) imin++;
  imax = 8;
  while (posP[imax] > dmax) imax--;

  for (ampP = 0, i = imin; i <= imax; i++) {
    if (corrP[i*npulse+NRS_CP] > SYNC_THRESHOLD &&
      ampP < corrP[i*npulse+NRS_CP])
    {
      ampP = corrP[i*npulse+NRS_CP];
      indP = i;
    }
  }

  /* search for negative peak */
  dmin = ref - 2*PCM_LENGTH - nTrack;
  dmax = ref - 2*PCM_LENGTH + nTrack;

  imin = 0;
  while (posN[imin] < dmin) imin++;
  imax = 8;
  while (posN[imax] > dmax) imax--;

  for (ampN = 0, i = imin; i <= imax; i++) {
    if (corrN[i*npulse+NRS_CP] > SYNC_THRESHOLD &&
      ampN < corrN[i*npulse+NRS_CP])
    {
      ampN = corrN[i*npulse+NRS_CP];
      indN = i;
    }
  }

  /* final tracking decision */
  if (npeaks == 0) {
    if ((
      posN[indN]-posP[indP] == PEAK_DIST_PN && ampP > 0 && ampN > 0 &&
      MIN(ampP, ampN) > MAX(ampP, ampN)/3) || (
      ampN > sync->amplitude[0]/2 &&
      ampP < ampN/6))
    {
      sync->delayMem = sync->delay;
      sync->delay    = posN[indN] - 4 - NRS_CP + PCM_LENGTH;
      sync->flag     = True;

      LogInfo("[PSAP] sync tracking detects sample slip! [%+1d sample(s)]",
        (sync->delay)-(sync->delayMem));
    }
    else if (
      ampP > sync->amplitude[0]/2 &&
      ampN < ampP/6)
    {
      sync->delayMem = sync->delay;
      sync->delay    = posP[indP] - 4 - NRS_CP + PCM_LENGTH + PEAK_DIST_PN;
      sync->flag     = True;

      LogInfo("[PSAP] sync tracking detects sample slip! [%+1d sample(s)]",
        (sync->delay)-(sync->delayMem));
    }
  }

  /* count consecutive unsuccessful tracking events */
  if (npeaks < 2 && sync->flag == False) {
    sync->checkCnt++;
  } else {
    sync->checkCnt = 0;
  }
}

/*============================================================================*/
/* FUNCTION: SyncFilter                                                       */
/*----------------------------------------------------------------------------*/
/* Description: sync filter implementation                                    */
/*                                                                            */
/* InOut:  SyncState*   sync    <-> sync struct                               */
/* In:     const Int16* pcm      -> input frame                               */
/*         Bool         invert   -> port inversion flag                       */
/*----------------------------------------------------------------------------*/

void SyncFilter(SyncState *sync, const Int16 *pcm, Bool invert)
{
  Int16 i, j, jmax, nframes;

  Int16 idxmax = syncIndexPreamble[SYNC_IDXLEN-1];
  Int16 offset = idxmax + PCM_LENGTH + 2*NRS_CP;
  Int16 npulse = 2*NRS_CP+1;
  Int32 sgnpcm = (invert == True) ? -1 : 1;

  /* set buffer pointers */
  Int32 *posP  = &sync->mem[offset +  0];
  Int32 *posN  = &sync->mem[offset +  9];
  Int32 *corrP = &sync->mem[offset + 18];
  Int32 *corrN = &sync->mem[offset + 27 + 9*2*NRS_CP];

  /* sync filter function */
  jmax = SYNC_IDXLEN-1;
  for (i = -idxmax; i < PCM_LENGTH; i++) {
    while (jmax > 0 && (syncIndexPreamble[jmax]+i) >= PCM_LENGTH) {
      jmax--;
    }
    j = jmax;
    while ((syncIndexPreamble[j]+i) >=0 && j >= 0) {
      sync->mem[i+idxmax+2*NRS_CP] +=
        sgnpcm*
        (Int32)pcm[syncIndexPreamble[j]+i]*
        (Int32)syncPulseForm[j%5]*
        (Int32)syncSequence[j/5];
      j--;
    }
  }
  /* shift all peaks by one slot */
  for (i = 1; i < 9; i++) {
    posP[i-1] = posP[i] - PCM_LENGTH;
    posN[i-1] = posN[i] - PCM_LENGTH;
    memcpy(&corrP[(i-1)*npulse], &corrP[i*npulse], sizeof(Int32)*npulse);
    memcpy(&corrN[(i-1)*npulse], &corrN[i*npulse], sizeof(Int32)*npulse);
  }
  memset(&corrP[8*npulse], 0, sizeof(Int32)*npulse);
  memset(&corrN[8*npulse], 0, sizeof(Int32)*npulse);

  /* search for maximum and minimum */
  for (i = 0; i < PCM_LENGTH; i++) {
    if (corrP[8*npulse+NRS_CP] < sync->mem[i+NRS_CP]) {
      memcpy(&corrP[8*npulse], &sync->mem[i], sizeof(Int32)*npulse);
      posP[8] = i+1;
    }
    if (corrN[8*npulse+NRS_CP] > sync->mem[i+NRS_CP]) {
      memcpy(&corrN[8*npulse], &sync->mem[i], sizeof(Int32)*npulse);
      posN[8] = i+1;
    }
  }
  for (i = 8*npulse; i < 8*npulse+npulse; i++) {
    corrN[i] = -corrN[i];
  }

  /* shift filter memory by one frame */
  nframes = (idxmax+2*NRS_CP)/PCM_LENGTH;

  for (i = 1; i <= nframes; i++) {
    memcpy(&sync->mem[(i-1)*PCM_LENGTH], &sync->mem[i*PCM_LENGTH],
      sizeof(Int32)*PCM_LENGTH);
  }
  memcpy(&sync->mem[nframes*PCM_LENGTH], &sync->mem[(nframes+1)*PCM_LENGTH],
    sizeof(Int32)*((idxmax+2*NRS_CP)%PCM_LENGTH));
  memset(&sync->mem[idxmax+2*NRS_CP], 0, sizeof(Int32)*PCM_LENGTH);

  /* compute global maximum amplitude */
  sync->amplitude[1] =
    MAX(sync->amplitude[1],
    MAX(corrP[8*npulse+NRS_CP], corrN[8*npulse+NRS_CP]));
}

/*============================================================================*/
/* UTILITY FUNCTION: SyncSubPut, SyncSubGet, SyncSubCpy                       */
/*----------------------------------------------------------------------------*/
/* InOut:  SyncState* sync  <-> sync struct                                   */
/* InOut:  SyncSub*   ssub  <-> sync subsystem                                */
/*----------------------------------------------------------------------------*/

void SyncSubPut(const SyncState *sync, SyncSub *ssub)
{
  Int16 i;
  for (i = 0; i <= 2*NRS_CP; i++) {
    ssub->shape[i] = sync->shape[i];
  }
  ssub->flag         = sync->flag;
  ssub->delay        = sync->delay;
  ssub->npeaks       = sync->npeaks;
  ssub->amplitude[0] = sync->amplitude[0];
  ssub->amplitude[1] = sync->amplitude[1];
}

void SyncSubGet(SyncState *sync, const SyncSub *ssub)
{
  Int16 i;
  for (i = 0; i <= 2*NRS_CP; i++) {
    sync->shape[i] = ssub->shape[i];
  }
  sync->flag         = ssub->flag;
  sync->delay        = ssub->delay;
  sync->npeaks       = ssub->npeaks;
  sync->amplitude[0] = ssub->amplitude[0];
  sync->amplitude[1] = ssub->amplitude[1];
}

void SyncSubCpy(const SyncSub *ssubIn, SyncSub *ssubOut)
{
  Int16 i;
  for (i = 0; i <= 2*NRS_CP; i++) {
    ssubOut->shape[i] = ssubIn->shape[i];
  }
  ssubOut->flag         = ssubIn->flag;
  ssubOut->delay        = ssubIn->delay;
  ssubOut->npeaks       = ssubIn->npeaks;
  ssubOut->amplitude[0] = ssubIn->amplitude[0];
  ssubOut->amplitude[1] = ssubIn->amplitude[1];
}

/*============================================================================*/
/* UTILITY FUNCTION: SyncSubRun                                               */
/*----------------------------------------------------------------------------*/
/* Description: sync peak evaluation                                          */
/*                                                                            */
/* InOut:  SyncSub*     ssub    <-> sync subsystem                            */
/* In:     const char*  caller   -> modem identification                      */
/*         const Int32* posP     -> positive peak positions                   */
/*         const Int32* corrP    -> positive peak correlation values          */
/*         const Int32* posN     -> negative peak positions                   */
/*         const Int32* corrN    -> negative peak correlation values          */
/*----------------------------------------------------------------------------*/

void SyncSubRun(SyncSub *ssub, const char *caller,
                const Int32 *posP, const Int32 *corrP,
                const Int32 *posN, const Int32 *corrN)
{
  Int16 i, tgtP, tgtN;
  Int16 p[5];

  Bool detP[5];
  Bool detN[5];

  /* initialize */
  Int16 npeaksP = 0;
  Int16 npeaksN = 0;
  Int16 npulse  = 2*NRS_CP+1;

  /* reset sync flag for peak detection */
  ssub->flag = False;

  /* frame numbers for peak positions */
  p[0] = PeakUpdate(posN, corrN, -2*PEAK_DIST_PN - PEAK_DIST_PP);
  p[1] = PeakUpdate(posP, corrP, -PEAK_DIST_PN - PEAK_DIST_PP);
  p[2] = PeakUpdate(posP, corrP, -PEAK_DIST_PN - PEAK_DIST_PP/2);
  p[3] = PeakUpdate(posP, corrP, -PEAK_DIST_PN);
  p[4] = 8;

  /* target delays */
  for (i = 0; i < 5; i++) {
    detP[i] = False;
    detN[i] = False;
  }
  tgtP = posP[p[3]] - 4 - PCM_LENGTH - NRS_CP + PEAK_DIST_PN;
  tgtN = posN[p[4]] - 4 - PCM_LENGTH - NRS_CP;

  /* check peak distances with respect to p[3] */
  if (posP[p[3]]-posP[p[1]] == PEAK_DIST_PP &&
    corrP[p[1]*npulse+NRS_CP] > SYNC_THRESHOLD &&
    corrP[p[3]*npulse+NRS_CP] > SYNC_THRESHOLD)
  {
    npeaksP = 2;
    detP[1] = detP[3] = True;

    if (posP[p[3]]-posN[p[0]] == PEAK_DIST_PN + PEAK_DIST_PP &&
      corrN[p[0]*npulse+NRS_CP] > SYNC_THRESHOLD)
    {
      npeaksP++;
      detP[0] = True;
    }
    if (posP[p[3]]-posP[p[2]] == PEAK_DIST_PP/2 &&
      corrP[p[2]*npulse+NRS_CP] > SYNC_THRESHOLD)
    {
      npeaksP++;
      detP[2] = True;
    }
    if (posN[p[4]]-posP[p[3]] == PEAK_DIST_PN &&
      corrN[p[4]*npulse+NRS_CP] > SYNC_THRESHOLD)
    {
      npeaksP++;
      detP[4] = True;
    }
  }

  /* check peak distances with respect to p[4] */
  if (posN[p[4]]-posN[p[0]] == PEAK_DIST_NN &&
    corrN[p[0]*npulse+NRS_CP] > SYNC_THRESHOLD &&
    corrN[p[4]*npulse+NRS_CP] > SYNC_THRESHOLD)
  {
    npeaksN = 2;
    detN[0] = detN[4] = True;

    if (posN[p[4]]-posP[p[1]] == PEAK_DIST_PN + PEAK_DIST_PP &&
      corrP[p[1]*npulse+NRS_CP] > SYNC_THRESHOLD)
    {
      npeaksN++;
      detN[1] = True;
    }
    if (posN[p[4]]-posP[p[2]] == PEAK_DIST_PN + PEAK_DIST_PP/2 &&
      corrP[p[2]*npulse+NRS_CP] > SYNC_THRESHOLD)
    {
      npeaksN++;
      detN[2] = True;
    }
    if (posN[p[4]]-posP[p[3]] == PEAK_DIST_PN &&
      corrP[p[3]*npulse+NRS_CP] > SYNC_THRESHOLD)
    {
      npeaksN++;
      detN[3] = True;
    }
  }

  /* evaluate detection result */
  if (npeaksP < npeaksN) {
    PeakCheck(ssub, caller, detN, p,
      corrP, corrN, corrN, p[0], p[4], npeaksN, tgtN);

    if (npeaksP > 0) {
      PeakCheck(ssub, caller, detP, p,
        corrP, corrN, corrP, p[1], p[3], npeaksP, tgtP);
    }
  }
  else {
    if (npeaksP > 0) {
      PeakCheck(ssub, caller, detP, p,
        corrP, corrN, corrP, p[1], p[3], npeaksP, tgtP);
    }
    if (npeaksN > 0) {
      PeakCheck(ssub, caller, detN, p,
        corrP, corrN, corrN, p[0], p[4], npeaksN, tgtN);
    }
  }
}

/*============================================================================*/
/* UTILITY FUNCTION: SyncSubChk                                               */
/*----------------------------------------------------------------------------*/
/* Description: sync peak evaluation for sync check                           */
/*                                                                            */
/* InOut:  SyncSub*     ssub    <-> sync subsystem                            */
/* In:     const char*  caller   -> modem identification                      */
/*         const Int32* posP     -> positive peaks positions                  */
/*         const Int32* corrP    -> positive peaks correlation values         */
/*         const Int32* posN     -> negative peaks positions                  */
/*         const Int32* corrN    -> negative peaks correlation values         */
/*         Int16        delay    -> reference delay                           */
/*----------------------------------------------------------------------------*/

void SyncSubChk(SyncSub *ssub, const char *caller,
                const Int32 *posP, const Int32 *corrP,
                const Int32 *posN, const Int32 *corrN, Int16 delay)
{
  Int16 p[5];
  Int16 q[5];
  Int16 ref = (delay + 4 + PCM_LENGTH + NRS_CP - 1)%PCM_LENGTH + 1;
  Int16 npulse = 2*NRS_CP+1;

  /* frame numbers for peak positions */
  q[0] = ref - 2*PEAK_DIST_PN - PEAK_DIST_PP;
  q[1] = ref - PEAK_DIST_PN - PEAK_DIST_PP;
  q[2] = ref - PEAK_DIST_PN - PEAK_DIST_PP/2;
  q[3] = ref - PEAK_DIST_PN;
  q[4] = ref;

  p[0] = 7 + q[0]/PCM_LENGTH;
  p[1] = 7 + q[1]/PCM_LENGTH;
  p[2] = 7 + q[2]/PCM_LENGTH;
  p[3] = 7 + q[3]/PCM_LENGTH;
  p[4] = 8;

  ssub->npeaksChk = 0;

  if (posN[p[0]] == q[0] && corrN[p[0]*npulse+NRS_CP] > SYNC_THRESHOLD) {
    ssub->npeaksChk++;
  }
  if (posP[p[1]] == q[1] && corrP[p[1]*npulse+NRS_CP] > SYNC_THRESHOLD) {
    ssub->npeaksChk++;
  }
  if (posP[p[2]] == q[2] && corrP[p[2]*npulse+NRS_CP] > SYNC_THRESHOLD) {
    ssub->npeaksChk++;
  }
  if (posP[p[3]] == q[3] && corrP[p[3]*npulse+NRS_CP] > SYNC_THRESHOLD) {
    ssub->npeaksChk++;
  }
  if (posN[p[4]] == q[4] && corrN[p[4]*npulse+NRS_CP] > SYNC_THRESHOLD) {
    ssub->npeaksChk++;
  }
}

/*============================================================================*/
/* UTILITY FUNCTION: PeakUpdate                                               */
/*----------------------------------------------------------------------------*/
/* Description: update sync peak position                                     */
/*                                                                            */
/* In:     const Int32* pos   -> vector of positions                          */
/*         const Int32* corr  -> vector of correlation values                 */
/*         Int16        dist  -> distance to be checked                       */
/* Return: Int16              <- updated peak position                        */
/*----------------------------------------------------------------------------*/

Int16 PeakUpdate(const Int32 *pos, const Int32 *corr, Int16 dist)
{
  Int16 npulse = 2*NRS_CP+1;
  Int16 i = dist + PCM_LENGTH;
  Int16 j = dist + 1;
  Int16 peak;

  if (pos[7+i/PCM_LENGTH] <= i) {
    if (pos[7+j/PCM_LENGTH] < j) {
      peak = 7+i/PCM_LENGTH;
    } else {
      peak = (corr[npulse*(7+j/PCM_LENGTH)+NRS_CP] <
        corr[npulse*(7+i/PCM_LENGTH)+NRS_CP]) ?
        7+i/PCM_LENGTH : 7+j/PCM_LENGTH;
    }
  } else {
    peak = (pos[7+j/PCM_LENGTH] < j) ? 0 : 7+j/PCM_LENGTH;
  }
  return peak;
}

/*============================================================================*/
/* UTILITY FUNCTION: PeakCheck                                                */
/*----------------------------------------------------------------------------*/
/* Description: check sync peaks                                              */
/*                                                                            */
/* InOut:  SyncSub*     ssub    <-> sync subsystem                            */
/* In:     const char*  caller   -> modem identification                      */
/*         const Bool*  pdet     -> vector of peak detection flags            */
/*         const Int16* p        -> vector of frame numbers                   */
/*         const Int32* corr(X)  -> vector of correlation values              */
/*         Int16        pos1     -> peak position 1                           */
/*         Int16        pos2     -> peak position 2                           */
/*         Int16        npeaks   -> number of detected peaks                  */
/*         Int16        delay    -> target delay if sync successful           */
/*----------------------------------------------------------------------------*/

void PeakCheck(SyncSub *ssub,
               const char *caller, const Bool *pdet, const Int16 *p,
               const Int32 *corrP, const Int32 *corrN, const Int32 *corr,
               Int16 pos1, Int16 pos2, Int16 npeaks, Int16 delay)
{
  Int16 i, cond1, cond2, jmax = 0, npulse = 2*NRS_CP+1;
  Int32 atmp, amax, ssum=0; 

  /* run only if sync has not yet been successful */
  if (ssub->flag == False) {
    atmp = corr[pos1*npulse+NRS_CP]/2 + corr[pos2*npulse+NRS_CP]/2;

    if (npeaks > MAX(ssub->npeaks, 3)) {
      ssub->flag = True;
    }
    else {
      amax = MAX(corr[pos2*npulse+NRS_CP], corr[pos1*npulse+NRS_CP]);
      for (i = 0; i < 9; i++) {
        if (corrP[i*npulse+NRS_CP] >= amax) jmax++;
      }
      if (jmax <= 3) {
        cond1 =
          ssub->amplitude[0] < atmp &&
          ssub->amplitude[1]/2 < atmp &&
          ssub->npeaks < npeaks &&
          MAX(corr[pos1*npulse+NRS_CP], corr[pos2*npulse+NRS_CP])/3 <
          MIN(corr[pos1*npulse+NRS_CP], corr[pos2*npulse+NRS_CP]);

        cond2 = ssub->amplitude[0] < atmp && npeaks >= MAX(ssub->npeaks, 4);

        if (cond1 || cond2) {
          ssub->flag = True;
        }
      }
    }
    if (ssub->flag == True) {
      /* compute normalized reference pulse shape */
      for (i = 0; i < npulse; i++) {
        ssub->shape[i] =
          pdet[0]*(corrN[p[0]*npulse+i] >> 3) +
          pdet[1]*(corrP[p[1]*npulse+i] >> 3) +
          pdet[2]*(corrP[p[2]*npulse+i] >> 3) +
          pdet[3]*(corrP[p[3]*npulse+i] >> 3) +
          pdet[4]*(corrN[p[4]*npulse+i] >> 3);
      }
      amax = 0;
      for (i = 0; i < npulse; i++) {
        amax += (ssub->shape[i] >> 16)*(ssub->shape[i] >> 16);
      }
      /* avoid division by zero */
      if ((amax = amax >> 12) == 0) {
        amax = 1;
      }
      for (i = 0; i < npulse; i++) {
        ssub->shape[i] =
          (ABS(ssub->shape[i] >> 16)*(ssub->shape[i] >> 16))/amax;
        ssum=ssum+ssub->shape[i]/npulse; 
      }

      if (ABS(ssum)<10) { 
        ssub->flag = False;
      }
      ssub->delay        = delay;
      ssub->npeaks       = npeaks;
      ssub->amplitude[0] = atmp;
      ssub->sign         = ssum;
    }
  }
}

/*============================================================================*/
/* UTILITY FUNCTION: ToneDetect                                               */
/*----------------------------------------------------------------------------*/
/* Description: tone detection at 500 Hz or 800 Hz                            */
/*                                                                            */
/* InOut:  SyncState*   sync  <-> sync struct                                 */
/* In:     const Int16* pcm    -> input frame                                 */
/*----------------------------------------------------------------------------*/

void ToneDetect(SyncState *sync, const Int16 *pcm)
{
  Int16 i;
  Int32 sumSin500 = 0;
  Int32 sumCos500 = 0;
  Int32 sumSin800 = 0;
  Int32 sumCos800 = 0;

  if (sync->memWakeup != NULL) {
    for (i = 0; i < NRF_SYNC; i++) {
      sync->memWakeup[2*i]   = sync->memWakeup[2*(i+1)];
      sync->memWakeup[2*i+1] = sync->memWakeup[2*(i+1)+1];
    }
    for (i = 0; i < PCM_LENGTH; i++) {
      sumSin500 += (wakeupSin500[i%16] >> 5)*pcm[i];
      sumCos500 += (wakeupCos500[i%16] >> 5)*pcm[i];
      sumSin800 += (wakeupSin800[i%10] >> 5)*pcm[i];
      sumCos800 += (wakeupCos800[i%10] >> 5)*pcm[i];
    }
    sync->memWakeup[2*NRF_SYNC] =
      (sumSin500 >> 18)*(sumSin500 >> 18) +
      (sumCos500 >> 18)*(sumCos500 >> 18);

    sync->memWakeup[2*NRF_SYNC+1] =
      (sumSin800 >> 18)*(sumSin800 >> 18) +
      (sumCos800 >> 18)*(sumCos800 >> 18);
  }
}
