/**
 * TestLab main driver: socket implementation.
 *
 * @copyright 2008, Research In Motion Ltd.
 *
 * @version $Id: //depot/main/dsp/ecall/c/testlab/testlab_sock.c#12 $
 */

#include <stdlib.h>

#include "rimtypes.h"
#include "ipc.h"
#include "error.h"
#include "option.h"
#include "testlab.h"

static char *options[] = {
  "--ul:%s@%s",
  "--ul@%s",
  "--ul",
  "--dl:%s@%s",
  "--dl@%s",
  "--dl",
  "--control:%s:%s:%s:%s@%s",
  "--control@%s",
  "--control",
  "--logpcm:%s",
  "--logpcm",
  "--log:%s",
  "--loghlack",
  "--ivspush",
  "--ackstop",
  "--startframe:%s",
  "--startframehlack:%s",
  "--totalframes:%s",
  "--totalframesack:%s",
  "--campaign:%s",
  "--invcodec:%s",
  "--invcodec",
  "-h",
  "--help",
  NULL
};

#define ARG_MAX_LENGTH  (200)
#define PORT_MAX_LENGTH (200)
#define HOST_MAX_LENGTH (200)
#define FILE_MAX_LENGTH (200)

static char p1[ARG_MAX_LENGTH];
static char p2[ARG_MAX_LENGTH];
static char p3[ARG_MAX_LENGTH];
static char p4[ARG_MAX_LENGTH];
static char p5[ARG_MAX_LENGTH];

static char ulPort  [PORT_MAX_LENGTH + 1] = "";
static char dlPort  [PORT_MAX_LENGTH + 1] = "";
static char ivsPort [PORT_MAX_LENGTH + 1] = "";
static char psapPort[PORT_MAX_LENGTH + 1] = "";
static char host    [HOST_MAX_LENGTH + 1] = "";

static char *pcm = NULL;
static char invCodec[4 + 1] = "none";
static char pcmFile[FILE_MAX_LENGTH + 1] = "";
static char logFile[FILE_MAX_LENGTH + 1] = "";
static char cpnFile[FILE_MAX_LENGTH + 1] = "";

static int logHlack = FALSE;
static int ivsPush  = FALSE;
static int ackStop  = FALSE;

static int mode            = 0;
static int startFrame      = 0;
static int startFrameHlack = 0;
static int totalFrames     = 0;
static int totalFramesAck  = 0;

static void Usage(char *msg, ...) {
  va_list args;
  va_start(args, msg);

  if (msg != NULL) {
    fprintf(stderr, "Error: ");
    vfprintf(stderr, msg, args);
    fprintf(stderr, "\n");
  }

  fprintf(stderr, "Usage is one of the following forms:\n");
  fprintf(stderr, "  testlab --ul[[:port]@host]\n");
  fprintf(stderr, "    Runs testlab as the UL channel processing task.\n");
  fprintf(stderr, "    Default port@host is 10000@localhost\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "  testlab --dl[[:port]@host]\n");
  fprintf(stderr, "    Runs testlab as the DL channel processing task.\n");
  fprintf(stderr, "    Default port@host is 10001@localhost\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "  testlab <options> --control[[:ul:dl:ivs:psap]@host]\n");
  fprintf(stderr, "    Runs testlab as the control processing task.\n");
  fprintf(stderr, "    Default ul:dl:ivs:psap@host is\n");
  fprintf(stderr, "    10000:10001:10002:10003@localhost\n");
  fprintf(stderr, "\n");
  fprintf(stderr, "    with <options> in:\n");
  fprintf(stderr, "  --logpcm[:fileBase]\n");
  fprintf(stderr, "    Logs UL in/out and DL in/out PCM samples.\n");
  fprintf(stderr, "    Default fileBase is pcm.\n");
  fprintf(stderr, "    PCM file name template: fileBase<src><index>.pcm\n");
  fprintf(stderr, "  --log:file\n");
  fprintf(stderr, "    Sets output log file to file.\n");
  fprintf(stderr, "    Default is results.txt.\n");
  fprintf(stderr, "    [stdout] and [stderr] can be used to send output to\n");
  fprintf(stderr, "    stdout and stderr streams.\n");
  fprintf(stderr, "  --loghlack\n");
  fprintf(stderr, "    Define to log HLACK reception result to output file.\n");
  fprintf(stderr, "    Default is FALSE.\n");
  fprintf(stderr, "  --ivspush\n");
  fprintf(stderr, "    Define to use IVS initiated signaling (push mode).\n");
  fprintf(stderr, "    Default is FALSE.\n");
  fprintf(stderr, "  --ackstop\n");
  fprintf(stderr, "    Define to stop control loop when ACK was received,\n");
  fprintf(stderr, "    which disables HLACK data reception.\n");
  fprintf(stderr, "    Default is FALSE.\n");
  fprintf(stderr, "  --startframe:n\n");
  fprintf(stderr, "    Set the number of frames to process before sending\n");
  fprintf(stderr, "    the MsgPsapTx signal to the PSAP.\n");
  fprintf(stderr, "    Default is 50 frames (1 s).\n");
  fprintf(stderr, "  --startframehlack:n\n");
  fprintf(stderr, "    Set the number of frames to process before sending\n");
  fprintf(stderr, "    the MsgPsapHlack signal during ACK transmission.\n");
  fprintf(stderr, "    Default is 10 frames (200 ms).\n");
  fprintf(stderr, "  --totalframes:n\n");
  fprintf(stderr, "    Set the number of frames to wait for the MsgPsapRx\n");
  fprintf(stderr, "    signal before declaring a transmission timeout.\n");
  fprintf(stderr, "    Default is 200*50 frames (200 s).\n");
  fprintf(stderr, "  --totalframesack:n\n");
  fprintf(stderr, "    Set the number of frames to process after reception\n");
  fprintf(stderr, "    of the MsgPsapRx signal.\n");
  fprintf(stderr, "    Default is 250 frames (5 s).\n");
  fprintf(stderr, "  --campaign:file\n");
  fprintf(stderr, "    Set the input campaign file to file.\n");
  fprintf(stderr, "    Default is campaign.txt\n");
  fprintf(stderr, "  --invcodec[:dir]\n");
  fprintf(stderr, "    Invert vocoder output signal (multiply by -1).\n");
  fprintf(stderr, "    Option: dir=ul or dir=dl, otherwise both directions.\n");
  fprintf(stderr, "    Default is FALSE.\n");

  va_end(args);
}

static int ProcessArgs(int argc, char *argv[]) {
  int ulCount = 0;
  int dlCount = 0;
  int controlCount = 0;

  while (argc-- > 0) {
    int n;

    if (strlen(*argv) >= ARG_MAX_LENGTH) {
      Usage("Argument too long <%s>", *argv);
      return FALSE;
    }

    n = option(*argv, options, p1, p2, p3, p4, p5);

    switch (n) {
      case 0:  /* --ul:%s@%s */
        strcpy(ulPort, p1);
        strcpy(host, p2);
        ulCount += 1;
        break;
      case 1:  /* --ul@%s */
        strcpy(host, p1);
        ulCount += 1;
        break;
      case 2:  /* --ul */
        ulCount += 1;
        break;
      case 3:  /* --dl:%s@%s */
        strcpy(dlPort, p1);
        strcpy(host, p2);
        dlCount += 1;
        break;
      case 4:  /* --dl@%s */
        strcpy(host, p1);
        dlCount += 1;
        break;
      case 5:  /* --dl */
        dlCount += 1;
        break;
      case 6:  /* --control:%s:%s:%s:%s@%s */
        strcpy(ulPort, p1);
        strcpy(dlPort, p2);
        strcpy(ivsPort, p3);
        strcpy(psapPort, p4);
        strcpy(host, p5);
        controlCount += 1;
        break;
      case 7:  /* --control@%s */
        strcpy(host, p1);
        controlCount += 1;
        break;
      case 8:  /* --control */
        controlCount += 1;
        break;
      case 9:  /* --logpcm:%s */
        strcpy(pcmFile, p1);
        pcm = pcmFile;
        break;
      case 10: /* --logpcm */
        pcm = pcmFile;
        break;
      case 11: /* --log:%s */
        strcpy(logFile, p1);
        break;
      case 12: /* --loghlack */
        logHlack = TRUE;
        break;
      case 13: /* --ivspush */
        ivsPush = TRUE;
        break;
      case 14: /* --ackstop */
        ackStop = TRUE;
        break;
      case 15: /* --startframe:%s */
        if (sscanf(p1, "%d", &startFrame) != 1) {
          Usage("Invalid startFrame count <%s>", p1);
          return FALSE;
        }
        break;
      case 16: /* --startframehlack:%s */
        if (sscanf(p1, "%d", &startFrameHlack) != 1) {
          Usage("Invalid startFrameHlack count <%s>", p1);
          return FALSE;
        }
        break;
      case 17: /* --totalframes:%s */
        if (sscanf(p1, "%d", &totalFrames) != 1) {
          Usage("Invalid totalFrames count <%s>", p1);
          return FALSE;
        }
        break;
      case 18: /* --totalframesack:%s */
        if (sscanf(p1, "%d", &totalFramesAck) != 1) {
          Usage("Invalid totalFramesAck count <%s>", p1);
          return FALSE;
        }
        break;
      case 19: /* --campaign:%s */
        strcpy(cpnFile, p1);
        break;
      case 20: /* --invcodec:%s */
        strncpy(invCodec, p1, 4);
        break;
      case 21: /* --invcodec */
        strcpy(invCodec, "both");
        break;
      case 22: /* -h */
      case 23: /* --help */
        Usage(NULL);
        return FALSE;
      default:
        Usage("Invalid option %s", *argv);
        return FALSE;
    }
    argv++;
  }
  if ((ulCount + dlCount + controlCount) < 1) {
    Usage("At least one of --ul --dl or --control must appear");
    return FALSE;
  }
  if ((ulCount + dlCount + controlCount) > 1) {
    Usage("At most one of --ul --dl or --control can appear");
    return FALSE;
  }
  if (ulCount      > 0) mode = 1;
  if (dlCount      > 0) mode = 2;
  if (controlCount > 0) mode = 3;

  if (strlen(ulPort)   <= 0) strcpy(ulPort,   "10000");
  if (strlen(dlPort)   <= 0) strcpy(dlPort,   "10001");
  if (strlen(ivsPort)  <= 0) strcpy(ivsPort,  "10002");
  if (strlen(psapPort) <= 0) strcpy(psapPort, "10003");
  if (strlen(host)     <= 0) strcpy(host,     "localhost");
  if (strlen(pcmFile)  <= 0) strcpy(pcmFile,  "pcm");
  if (strlen(logFile)  <= 0) strcpy(logFile,  "results.txt");
  if (strlen(cpnFile)  <= 0) strcpy(cpnFile,  "campaign.txt");
  if (startFrame       == 0) startFrame      = 50;
  if (startFrameHlack  == 0) startFrameHlack = 10;
  if (totalFrames      == 0) totalFrames     = 10000;
  if (totalFramesAck   == 0) totalFramesAck  = 250;

  return TRUE;
}

static void Run(BOOL (*run)(void))
{
  IpcStart();
  while ((*run)());
  IpcStop();
}

int main(int argc, char *argv[]) {
  if (ProcessArgs(argc-1, &argv[1])) {
    LogInit(stderr);
    IpcInit();
    switch (mode) {
      case 1:
        UlChannelInit(host, ulPort);
        Run(UlChannelRun);
        UlChannelFinal();
        break;
      case 2:
        DlChannelInit(host, dlPort);
        Run(DlChannelRun);
        DlChannelFinal();
        break;
      case 3:
        ControlInit(host, ulPort, dlPort, ivsPort, psapPort,
          pcm, cpnFile, logFile, invCodec, logHlack, ivsPush, ackStop,
          totalFrames + startFrame, totalFramesAck,
          startFrame, startFrameHlack);
        LogInfo("Waiting for UL/DL/IVS/PSAP to connect");
        Run(ControlRun);
        break;
      default:
        return 0;
    }
    IpcFinal();
    return 0;
  }

  return -1;
}
